-- New sorting buttons for the invite lists
-- 23

local core = CalendarExtras

local module = core.NewModule("Sorters")

 --[[ -------- Locals ------- ]]--

local overrideSort, sortReverse
local listIsDirty

local roles -- This will point to the roles-table in the savedvars file after OnInit

-- Todo: Maybe add hybrid to the list... should've made these into icons in the first place
local roleIndexToName = {
	"Tank", -- [1]
	"Heal", -- [2]
	"DPS",  -- [3]
	"?"     -- [4]
}
local roleColors = {
	{ r = 0.78, g = 0.61, b = 0.43 }, -- Tank
	{ r = 0.1,  g = 1,    b = 0.1 },  -- Healer
	{ r = 1,    g = 0.1,  b = 0.1 },  -- DPS
	NORMAL_FONT_COLOR                 -- No role
}
local knownRolesPerClass = {
	MAGE    = 3,
	HUNTER  = 3,
	ROGUE   = 3,
	WARLOCK = 3
}

local roleContextMenu

local massSelectModule

 --[[ ----------------------- ]]--

local SetRole, GetRole, SortList

function SetRole(name, role)
	if not name or not role then return end
	roles[name] = role ~= 4 and role or nil
	if overrideSort then
		wipe(module.reSortedList)
		SortList()
	end
end

function GetRole(name, class)
	if name then
		if roles[name] then
			return roles[name]
		end
	end
	if class and knownRolesPerClass[class] then
		return knownRolesPerClass[class]
	end
	return 4
end

do -- function SortList()
	local statusTable, roleTable, inviteIndexTable = {}, {}, {}

	-- Need to move the status codes around so they sort in more sensible way
	local statusSwap = {
		3, 5, 2, 6, 1, 4
	}

	local function Sorter(name1, name2)
		if roleTable[name1] == roleTable[name2] then
			if statusTable[name1] == statusTable[name2] then
				return name1 < name2
			else
				return statusTable[name1] > statusTable[name2]
			end
		else
			return roleTable[name1] < roleTable[name2]
		end
	end

	function SortList()
		local t, i = {}, 1
		while true do
			local name, level, _, class, inviteStatus = CalendarEventGetInvite(i)
			if not name then break end
			t[i] = name
			statusTable[name] = statusSwap[inviteStatus]
			roleTable[name] = GetRole(name, class)
			inviteIndexTable[name] = i
			i = i + 1
		end
		table.sort(t, Sorter)
		for k, v in ipairs(t) do
			module.reSortedList[k] = inviteIndexTable[v]
		end
		wipe(statusTable)
		wipe(roleTable)
		wipe(inviteIndexTable)
	end
end

-- CalendarEventGetInvite() override, gets the real player index from our reSortedList table
local function CalendarEventGetInviteOverride(...)
	if overrideSort then
		local index = ...
		if sortReverse then
			index = #module.reSortedList - index + 1
		end
		if module.reSortedList[index] then
			return CalendarEventGetInvite(module.reSortedList[index])
		end
	end
	return CalendarEventGetInvite(...)
end

local function CalendarEventGetInviteSortCriterionOverride()
	if overrideSort then
		return overrideSort, sortReverse
	else
		return CalendarEventGetInviteSortCriterion()
	end
end

local function NewSorterOnClick(updateFunc, self, button, down)
	if #module.reSortedList == 0 then
		SortList()
	end
	overrideSort = "role"
	if sortReverse == nil then
		sortReverse = false
	else
		sortReverse = not sortReverse
	end
	PlaySound("igMainMenuOptionCheckBoxOn");
	CalendarContextMenu_Hide(CalendarViewEventInviteContextMenu_Initialize)
	CalendarContextMenu_Hide(CalendarCreateEventInviteContextMenu_Initialize)
	updateFunc()
	core.callbacks:Fire("CustomResorted")
end

local function UpdateInviteListButtons(_, frame)
	local offset = HybridScrollFrame_GetOffset(frame)
	local inviteIndex, name, class, role, roleButton, c, _
	for k, button in ipairs(frame.buttons) do
		-- Wait a sec... aren't I going in circles with this inviteIndex malarkey?
		inviteIndex = k + offset
		name, _, _, class = CalendarEventGetInviteOverride(inviteIndex)
		if not name then break end
		roleButton = _G[button:GetName().."Role"]
		role = GetRole(name, class)
		c = roleColors[role]
		roleButton:SetText(roleIndexToName[role])
		roleButton:SetTextColor(c.r, c.g, c.b)
		if overrideSort then
			if sortReverse then inviteIndex = #module.reSortedList - inviteIndex + 1 end
			button.inviteIndex = module.reSortedList[inviteIndex]
		end
	end
end

local function ClearOverride()
	wipe(module.reSortedList)
	overrideSort = nil
	sortReverse = nil
	listIsDirty = nil
end

local function CleanUp()
	if listIsDirty then
		listIsDirty = nil
		wipe(module.reSortedList)
		SortList()
	end
end

function module.OnInitialize()
	-- Set up our roles database
	if not core.db.roles then
		core.db.rolesDBver = 1 -- Think I screwed up when I didn't add hybirds, might have to update the tables at some point, so lets assign a version to it
		core.db.roles = {}
	end
	roles = core.db.roles

	module.reSortedList = {}

	if core.modules["Mass Select"] then
		massSelectModule = core.modules["Mass Select"]
	end

	if core.modules.Export then
		core.modules.Export.AddFormatToken("ROLE", "Role", function(name, class) return roleIndexToName[GetRole(name, class)] end)
	end

	-- Create the new sorting buttons
	-- Todo: The direction arrow is overlapped by the Role text for some reason
	local viewBtn = CreateFrame("Button", "CalendarViewEventInviteListRoleSortButton", CalendarViewEventInviteList, "CalendarEventInviteSortButtonTemplate")
	local createBtn = CreateFrame("Button", "CalendarCreateEventInviteListRoleSortButton", CalendarCreateEventInviteList, "CalendarEventInviteSortButtonTemplate")

	viewBtn:SetText("Role")
	viewBtn:SetPoint("BOTTOM", CalendarViewEventInviteList, "TOP", 3, 0)
	viewBtn.criterion = "role"

	createBtn:SetText("Role")
	createBtn:SetPoint("BOTTOM", CalendarCreateEventInviteList, "TOP", 3, 0)
	createBtn.criterion = "role"

	local width = viewBtn:GetTextWidth() + _G["CalendarViewEventInviteListRoleSortButtonDirection"]:GetWidth()
	viewBtn:SetWidth(width)
	createBtn:SetWidth(width)

	viewBtn:SetScript("OnClick", function(...) NewSorterOnClick(CalendarViewEventInviteList_Update, ...) end)
	createBtn:SetScript("OnClick", function(...) NewSorterOnClick(CalendarCreateEventInviteList_Update, ...) end)

	CalendarViewEventInviteList.sortButtons["role"] = viewBtn
	CalendarCreateEventInviteList.sortButtons["role"] = createBtn

	-- This is kinda stupid, would be better to modify the templates, but I'm not sure how to do that before the frames get created, if that's even possible
	local function Modify(button)
		local buttonName = button:GetName()
		local --[[name,]] class, status, role = --[[_G[buttonName.."Name"],]] _G[buttonName.."Class"], _G[buttonName.."Status"], button:CreateFontString(buttonName.."Role", "BORDER", "GameFontNormalSmall")

		class:SetWidth(70)
		class:ClearAllPoints()
		class:SetPoint("RIGHT", -110, 0)
		class:SetJustifyH("RIGHT")

		status:ClearAllPoints()
		status:SetPoint("RIGHT", -35, 0)
		status:SetJustifyH("RIGHT")

		role:SetWidth(30)
		role:SetJustifyH("RIGHT")
		role:SetPoint("RIGHT")
	end
	for _, button in ipairs(CalendarViewEventInviteList.scrollFrame.buttons) do
		Modify(button)
	end
	for _, button in ipairs(CalendarCreateEventInviteList.scrollFrame.buttons) do
		Modify(button)
	end

	module:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST", function()
		if overrideSort then
			listIsDirty = true
		end
	end)

	-- Clear everything before lists update after event is opened
	core.RegisterCallback(module, "OpenEvent", ClearOverride)

	-- Do cleanups before list is updated if needed
	core.RegisterCallback(module, "Pre_Update", CleanUp)

	-- Go through the buttons after list is updated by the game
	core.RegisterCallback(module, "Pre_Post_Update", UpdateInviteListButtons)

	-- Clear everything when list is resorted
	module:Hook("CalendarEventInviteSortButton_OnClick", ClearOverride, true)

	 -- Hook the sort button functions
	module:RawHook("CalendarEventInviteList_AnchorSortButtons", true)

	-- Environment proxies for the sccrollframe update functions, so they get updated properly even when they're sorted by my own functions. Shamelessly copied from Haste's iiWatcher iiRC.
	local ScrollFrame_Update_env = setmetatable({CalendarEventGetInvite = CalendarEventGetInviteOverride}, {__index = _G})
	setfenv(core.hooks.CalendarViewEventInviteListScrollFrame_Update, ScrollFrame_Update_env)
	setfenv(core.hooks.CalendarCreateEventInviteListScrollFrame_Update, ScrollFrame_Update_env)

	-- Env proxy for the CalendarEventInviteList_UpdateSortButtons
	local SortButtonUpdate_env = setmetatable({CalendarEventGetInviteSortCriterion = CalendarEventGetInviteSortCriterionOverride}, {__index = _G})
	setfenv(CalendarEventInviteList_UpdateSortButtons, SortButtonUpdate_env)
end

 --[[ -------- Hooks -------- ]]--

function module:CalendarEventInviteList_AnchorSortButtons(inviteList)
	local scrollFrame = inviteList.scrollFrame
	if not scrollFrame or not scrollFrame.buttons or not scrollFrame.buttons[1] then return end
	local inviteButtonName = scrollFrame.buttons[1]:GetName()

	local nameSortButton = inviteList.sortButtons.name
	if inviteList.partyMode then
		nameSortButton:SetPoint("LEFT", _G[inviteButtonName.."Name"], "LEFT")
	else
		nameSortButton:SetPoint("LEFT", _G[inviteButtonName.."PartyIcon"], "LEFT")
	end

	local classSortButton = inviteList.sortButtons.class
	classSortButton:SetPoint("RIGHT", _G[inviteButtonName.."Class"], "RIGHT")

	local statusSortButton = inviteList.sortButtons.status
	statusSortButton:SetPoint("RIGHT", _G[inviteButtonName.."Status"], "RIGHT")

	local roleSortButton = inviteList.sortButtons.role
	roleSortButton:SetPoint("RIGHT", _G[inviteButtonName.."Role"], "RIGHT")
end

 --[[ ------ Context Menu ------ ]]--

do
	roleContextMenu = CreateFrame("Frame", "CalendarRoleContextMenu", CalendarContextMenu, "UIMenuTemplate")
	roleContextMenu:Hide()
	roleContextMenu:SetFrameStrata("FULLSCREEN")
	roleContextMenu:EnableMouse(true)
	roleContextMenu:SetScript("OnShow", function(self)
		self:Initialize()
	end)
	roleContextMenu:SetScript("OnEvent", function(self, event)
		if event == "CALENDAR_UPDATE_EVENT" then
			if self:IsShown() then
				self:Initialize()
			end
		end
	end)
	roleContextMenu:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b)
	roleContextMenu:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b)
	roleContextMenu:RegisterEvent("CALENDAR_UPDATE_EVENT")
	roleContextMenu.parentMenu = "CalendarContextMenu"

	function roleContextMenu:Initialize()
		UIMenu_Initialize(self)

		local name = CalendarEventGetInvite(CalendarContextMenu.inviteButton.inviteIndex)

		for k, v in ipairs(roleIndexToName) do
			if v == "?" then
				UIMenu_AddButton(self)
			end
			UIMenu_AddButton(
				self,                   -- menu
				v == "Heal" and "Healer" or v == "?" and "Clear" or v, -- text
				nil,                    -- shortcut
				self.SetRole,           -- func
				nil,                    -- nested
				{name = name, role = k} -- value
			)
		end

		return UIMenu_FinishInitializing(self)
	end

	function roleContextMenu:SetRole()
		-- If the mass-select module is there and loaded, set the roles using it instead
		if massSelectModule then
			massSelectModule.RunOnSelected(SetRole, true, self.value.role)
		else
			SetRole(self.value.name, self.value.role)
		end

		-- Todo: Don't we need to mark the list dirty now?
		core.UpdateOpenList()
		CalendarContextMenu_Hide()
	end

	-- Hook the contex-menu generator to add in our role menu
	function module:CalendarCreateEventInviteContextMenu_Initialize(...)
		self.hooks.CalendarCreateEventInviteContextMenu_Initialize(...)

		local menu = ...
		UIMenu_AddButton(menu)
		UIMenu_AddButton(menu, "Set role", nil, nil, "CalendarRoleContextMenu")

		return UIMenu_FinishInitializing(menu)
	end
	module:RawHook("CalendarCreateEventInviteContextMenu_Initialize", true)

	-- Hook the contex-menu generator to add in our role menu. This doesn't normally open, so lets just replace the whole function with our own
	function module:CalendarViewEventInviteContextMenu_Initialize(...)
		local menu, inviteButton = ...

		UIMenu_Initialize(menu)
		CalendarInviteContextMenu_UnlockHighlights()
		menu.inviteButton = inviteButton
		inviteButton:LockHighlight()
		UIMenu_AddButton(menu, "Set role", nil, nil, "CalendarRoleContextMenu")

		return UIMenu_FinishInitializing(menu)
	end
	module:RawHook("CalendarViewEventInviteContextMenu_Initialize", true)

	-- Hook the viewevent onclick function to force our right-click menu open for role-setting.
	function module:CalendarViewEventInviteListButton_OnClick(self, button)
		if button == "LeftButton" then
			CalendarContextMenu_Hide()
		elseif button == "RightButton" then
			if CalendarContextMenu.inviteButton ~= self then
				CalendarContextMenu_Show(self, CalendarViewEventInviteContextMenu_Initialize, "cursor", 3, -3, self)
			else
				CalendarContextMenu_Toggle(self, CalendarViewEventInviteContextMenu_Initialize, "cursor", 3, -3, self)
			end
		end
		PlaySound("igMainMenuOptionCheckBoxOn")
	end
	module:RawHook("CalendarViewEventInviteListButton_OnClick", true)

	-- Hook to hide our role-contextmenu
	module:Hook("CalendarContextMenu_OnHide", function() CalendarRoleContextMenu:Hide() end, true)
end